"""
Manage GlusterFS pool.
"""

import logging

import salt.utils.cloud as suc
import salt.utils.network
from salt.exceptions import SaltCloudException

log = logging.getLogger(__name__)

RESULT_CODES = [
    "Peer {0} added successfully.",
    "Probe on localhost not needed",
    "Host {0} is already in the peer group",
    "Host {0} is already part of another cluster",
    "Volume on {0} conflicts with existing volumes",
    "UUID of {0} is the same as local uuid",
    '{0} responded with "unknown peer". This could happen if {0} doesn\'t have'
    " localhost defined",
    "Failed to add peer. Information on {0}'s logs",
    "Cluster quorum is not met. Changing peers is not allowed.",
    "Failed to update list of missed snapshots from {0}",
    "Conflict comparing list of snapshots from {0}",
    "Peer is already being detached from cluster.",
]


def __virtual__():
    """
    Only load this module if the gluster command exists
    """
    if "glusterfs.list_volumes" in __salt__:
        return "glusterfs"
    return (False, "glusterfs module could not be loaded")


def peered(name):
    """
    Check if node is peered.

    name
        The remote host with which to peer.

    .. code-block:: yaml

        peer-cluster:
          glusterfs.peered:
            - name: two

        peer-clusters:
          glusterfs.peered:
            - names:
              - one
              - two
              - three
              - four
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": False}

    try:
        suc.check_name(name, "a-zA-Z0-9._-")
    except SaltCloudException:
        ret["comment"] = "Invalid characters in peer name."
        return ret

    # Check if the name resolves to one of this minion IP addresses
    name_ips = salt.utils.network.host_to_ips(name)
    if name_ips is not None:
        # if it is None, it means resolution fails, let's not hide
        # it from the user.
        this_ips = set(salt.utils.network.ip_addrs())
        this_ips.update(salt.utils.network.ip_addrs6())
        if any(
            salt.utils.network.is_loopback(addr) for addr in name_ips
        ) or this_ips.intersection(name_ips):
            ret["result"] = True
            ret["comment"] = "Peering with localhost is not needed"
            return ret

    peers = __salt__["glusterfs.peer_status"]()

    if peers and any(name in v["hostnames"] for v in peers.values()):
        ret["result"] = True
        ret["comment"] = f"Host {name} already peered"
        return ret

    if __opts__["test"]:
        ret["comment"] = f"Peer {name} will be added."
        ret["result"] = None
        return ret

    if not __salt__["glusterfs.peer"](name):
        ret["comment"] = "Failed to peer with {}, please check logs for errors".format(
            name
        )
        return ret

    # Double check that the action succeeded
    newpeers = __salt__["glusterfs.peer_status"]()
    if newpeers and any(name in v["hostnames"] for v in newpeers.values()):
        ret["result"] = True
        ret["comment"] = f"Host {name} successfully peered"
        ret["changes"] = {"new": newpeers, "old": peers}
    else:
        ret["comment"] = (
            "Host {} was successfully peered but did not appear in the list of peers".format(
                name
            )
        )
    return ret


def volume_present(
    name,
    bricks,
    stripe=False,
    replica=False,
    device_vg=False,
    transport="tcp",
    start=False,
    force=False,
    arbiter=False,
):
    """
    Ensure that the volume exists

    name
        name of the volume

    bricks
        list of brick paths

    replica
        replica count for volume

    arbiter
        use every third brick as arbiter (metadata only)

        .. versionadded:: 2019.2.0

    start
        ensure that the volume is also started

    .. code-block:: yaml

        myvolume:
          glusterfs.volume_present:
            - bricks:
                - host1:/srv/gluster/drive1
                - host2:/srv/gluster/drive2

        Replicated Volume:
          glusterfs.volume_present:
            - name: volume2
            - bricks:
              - host1:/srv/gluster/drive2
              - host2:/srv/gluster/drive3
            - replica: 2
            - start: True

        Replicated Volume with arbiter brick:
          glusterfs.volume_present:
            - name: volume3
            - bricks:
              - host1:/srv/gluster/drive2
              - host2:/srv/gluster/drive3
              - host3:/srv/gluster/drive4
            - replica: 3
            - arbiter: True
            - start: True

    """
    ret = {"name": name, "changes": {}, "comment": "", "result": False}

    if suc.check_name(name, "a-zA-Z0-9._-"):
        ret["comment"] = "Invalid characters in volume name."
        return ret

    volumes = __salt__["glusterfs.list_volumes"]()
    if name not in volumes:
        if __opts__["test"]:
            comment = f"Volume {name} will be created"
            if start:
                comment += " and started"
            ret["comment"] = comment
            ret["result"] = None
            return ret

        vol_created = __salt__["glusterfs.create_volume"](
            name, bricks, stripe, replica, device_vg, transport, start, force, arbiter
        )

        if not vol_created:
            ret["comment"] = f"Creation of volume {name} failed"
            return ret
        old_volumes = volumes
        volumes = __salt__["glusterfs.list_volumes"]()
        if name in volumes:
            ret["changes"] = {"new": volumes, "old": old_volumes}
            ret["comment"] = f"Volume {name} is created"

    else:
        ret["comment"] = f"Volume {name} already exists"

    if start:
        if __opts__["test"]:
            # volume already exists
            ret["comment"] = ret["comment"] + " and will be started"
            ret["result"] = None
            return ret
        if int(__salt__["glusterfs.info"]()[name]["status"]) == 1:
            ret["result"] = True
            ret["comment"] = ret["comment"] + " and is started"
        else:
            vol_started = __salt__["glusterfs.start_volume"](name)
            if vol_started:
                ret["result"] = True
                ret["comment"] = ret["comment"] + " and is now started"
                if not ret["changes"]:
                    ret["changes"] = {"new": "started", "old": "stopped"}
            else:
                ret["comment"] = (
                    ret["comment"]
                    + " but failed to start. Check logs for further information"
                )
                return ret

    if __opts__["test"]:
        ret["result"] = None
    else:
        ret["result"] = True
    return ret


def started(name):
    """
    Check if volume has been started

    name
        name of the volume

    .. code-block:: yaml

        mycluster:
          glusterfs.started: []
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": False}

    volinfo = __salt__["glusterfs.info"]()
    if name not in volinfo:
        ret["result"] = False
        ret["comment"] = f"Volume {name} does not exist"
        return ret

    if int(volinfo[name]["status"]) == 1:
        ret["comment"] = f"Volume {name} is already started"
        ret["result"] = True
        return ret
    elif __opts__["test"]:
        ret["comment"] = f"Volume {name} will be started"
        ret["result"] = None
        return ret

    vol_started = __salt__["glusterfs.start_volume"](name)
    if vol_started:
        ret["result"] = True
        ret["comment"] = f"Volume {name} is started"
        ret["change"] = {"new": "started", "old": "stopped"}
    else:
        ret["result"] = False
        ret["comment"] = f"Failed to start volume {name}"

    return ret


def add_volume_bricks(name, bricks):
    """
    Add brick(s) to an existing volume

    name
        Volume name

    bricks
        List of bricks to add to the volume

    .. code-block:: yaml

        myvolume:
          glusterfs.add_volume_bricks:
            - bricks:
                - host1:/srv/gluster/drive1
                - host2:/srv/gluster/drive2

        Replicated Volume:
          glusterfs.add_volume_bricks:
            - name: volume2
            - bricks:
              - host1:/srv/gluster/drive2
              - host2:/srv/gluster/drive3
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": False}

    volinfo = __salt__["glusterfs.info"]()
    if name not in volinfo:
        ret["comment"] = f"Volume {name} does not exist"
        return ret

    if int(volinfo[name]["status"]) != 1:
        ret["comment"] = f"Volume {name} is not started"
        return ret

    current_bricks = [brick["path"] for brick in volinfo[name]["bricks"].values()]
    if not set(bricks) - set(current_bricks):
        ret["result"] = True
        ret["comment"] = f"Bricks already added in volume {name}"
        return ret

    bricks_added = __salt__["glusterfs.add_volume_bricks"](name, bricks)
    if bricks_added:
        ret["result"] = True
        ret["comment"] = f"Bricks successfully added to volume {name}"
        new_bricks = [
            brick["path"]
            for brick in __salt__["glusterfs.info"]()[name]["bricks"].values()
        ]
        ret["changes"] = {"new": new_bricks, "old": current_bricks}
        return ret

    ret["comment"] = f"Adding bricks to volume {name} failed"
    return ret


def op_version(name, version):
    """
    .. versionadded:: 2019.2.0

    Add brick(s) to an existing volume

    name
        Volume name

    version
        Version to which the cluster.op-version should be set

    .. code-block:: yaml

        myvolume:
          glusterfs.op_version:
            - name: volume1
            - version: 30707
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": False}

    try:
        current = int(__salt__["glusterfs.get_op_version"](name))
    except TypeError:
        ret["result"] = False
        ret["comment"] = __salt__["glusterfs.get_op_version"](name)[1]
        return ret

    if current == version:
        ret["comment"] = "Glusterfs cluster.op-version for {} already set to {}".format(
            name, version
        )
        ret["result"] = True
        return ret
    elif __opts__["test"]:
        ret["comment"] = (
            "An attempt would be made to set the cluster.op-version for {} to {}.".format(
                name, version
            )
        )
        ret["result"] = None
        return ret

    result = __salt__["glusterfs.set_op_version"](version)

    if result[0] is False:
        ret["comment"] = result[1]
        return ret

    ret["comment"] = result
    ret["changes"] = {"old": current, "new": version}
    ret["result"] = True
    return ret


def max_op_version(name):
    """
    .. versionadded:: 2019.2.0

    Add brick(s) to an existing volume

    name
        Volume name

    .. code-block:: yaml

        myvolume:
          glusterfs.max_op_version:
            - name: volume1
            - version: 30707
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": False}

    try:
        current = int(__salt__["glusterfs.get_op_version"](name))
    except TypeError:
        ret["result"] = False
        ret["comment"] = __salt__["glusterfs.get_op_version"](name)[1]
        return ret

    try:
        max_version = int(__salt__["glusterfs.get_max_op_version"]())
    except TypeError:
        ret["result"] = False
        ret["comment"] = __salt__["glusterfs.get_max_op_version"]()[1]
        return ret

    if current == max_version:
        ret["comment"] = (
            "The cluster.op-version is already set to the cluster.max-op-version of {}".format(
                current
            )
        )
        ret["result"] = True
        return ret
    elif __opts__["test"]:
        ret["comment"] = (
            "An attempt would be made to set the cluster.op-version to {}.".format(
                max_version
            )
        )
        ret["result"] = None
        return ret

    result = __salt__["glusterfs.set_op_version"](max_version)

    if result[0] is False:
        ret["comment"] = result[1]
        return ret

    ret["comment"] = result
    ret["changes"] = {"old": current, "new": max_version}
    ret["result"] = True
    return ret



"""
Unit tests for salt.utils.crypt.py
"""

import pytest

import salt.utils.crypt


@pytest.fixture
def pub_key_data():
    return [
        "-----BEGIN PUBLIC KEY-----",
        "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAyc9ehbU4J2uzPZZCEw8K",
        "5URYcKSUh0h/c6m9PR2kRFbXkHcSnpkWX+LCuFKQ5iF2+0rVn9pO/94rL5zAQ6DU",
        "lucqk9EvamSk+TjHh3Ps/HdSxxVbkLk3nglVJrDgENxnAz+Kp+OSNfI2uhhzJiu1",
        "Dhn86Wb46eu7EFYeJ+7z9+29UXuCiMIUL5sRx3Xy37gpiD4Z+JVtoBNx1MKJ4MqB",
        "24ZXsvtEyrCmuLwhKCiQqvNx91CkyIL+sfMoHDSf7sLwl1CuCEgny7EV7bJpoNzN",
        "ZFKggcJCopfzLWDijF5A5OOvvvFrr/rYjW79LkGviWTzJrBPNgoD01zWIlzJfLdh",
        "ywIDAQAB",
        "-----END PUBLIC KEY-----",
    ]


@pytest.mark.parametrize("line_ending", ["\n", "\r\n"])
def test_pem_finger_file_line_endings(tmp_path, pub_key_data, line_ending):
    key_file = tmp_path / "master_crlf.pub"
    key_file.write_bytes(line_ending.join(pub_key_data).encode("utf-8"))
    finger = salt.utils.crypt.pem_finger(path=str(key_file))
    assert (
        finger
        == "9b:42:66:92:8a:d1:b9:27:42:e0:6d:f3:12:c9:74:74:b0:e0:0e:42:83:87:62:ad:95:49:9d:6f:8e:d0:ed:35"
    )


@pytest.mark.parametrize("key", [b"123abc", "123abc"])
def test_pem_finger_key(key):
    finger = salt.utils.crypt.pem_finger(key=key)
    assert (
        finger
        == "dd:13:0a:84:9d:7b:29:e5:54:1b:05:d2:f7:f8:6a:4a:cd:4f:1e:c5:98:c1:c9:43:87:83:f5:6b:c4:f0:ff:80"
    )


def test_pem_finger_sha512():
    finger = salt.utils.crypt.pem_finger(key="123abc", sum_type="sha512")
    assert (
        finger
        == "7b:6a:d7:9b:34:6f:b6:95:12:75:34:39:48:e1:3c:1b:4e:bc:a8:2a:54:52:a6:c5:d1:56:84:37:7f:09:6c:a9:"
        "27:50:6a:23:a8:47:e6:e0:46:06:13:99:63:1b:16:fc:28:20:c8:b0:e0:2d:0e:a8:7a:a5:a2:03:a7:7c:2a:7e"
    )



"""
Salt compatibility code
"""

# pylint: disable=unused-import
import sys

# pragma: no cover

# The ipaddress module included in Salt is from Python 3.9.5.
# When running from Py3.9.5+ use the standard library module, use ours otherwise
if sys.version_info >= (3, 9, 5):
    import ipaddress
else:
    import salt.ext.ipaddress as ipaddress

# importlib_metadata before version 3.3.0 does not include the functionality we need.
try:
    import importlib_metadata

    importlib_metadata_version = [
        int(part)
        for part in importlib_metadata.version("importlib_metadata").split(".")
        if part.isdigit()
    ]
    if tuple(importlib_metadata_version) < (3, 3, 0):
        # Use the vendored importlib_metadata
        import salt.ext.importlib_metadata as importlib_metadata
except ImportError:
    # Use the vendored importlib_metadata
    import salt.ext.importlib_metadata as importlib_metadata



"""
Installer support for macOS.

Installer is the native .pkg/.mpkg package manager for macOS.
"""

import os.path
import urllib

import salt.utils.itertools
import salt.utils.mac_utils
import salt.utils.path
import salt.utils.platform
from salt.exceptions import SaltInvocationError

# Don't shadow built-in's.
__func_alias__ = {"list_": "list"}

__virtualname__ = "pkgutil"


def __virtual__():
    if not salt.utils.platform.is_darwin():
        return (False, "Only available on Mac OS systems")

    if not salt.utils.path.which("pkgutil"):
        return (False, "Missing pkgutil binary")

    return __virtualname__


def list_():
    """
    List the installed packages.

    :return: A list of installed packages
    :rtype: list

    CLI Example:

    .. code-block:: bash

        salt '*' pkgutil.list
    """
    cmd = "pkgutil --pkgs"
    ret = salt.utils.mac_utils.execute_return_result(cmd)
    return ret.splitlines()


def is_installed(package_id):
    """
    Returns whether a given package id is installed.

    :return: True if installed, otherwise False
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' pkgutil.is_installed com.apple.pkg.gcc4.2Leo
    """
    return package_id in list_()


def _install_from_path(path):
    """
    Internal function to install a package from the given path
    """
    if not os.path.exists(path):
        msg = f"File not found: {path}"
        raise SaltInvocationError(msg)

    cmd = f'installer -pkg "{path}" -target /'
    return salt.utils.mac_utils.execute_return_success(cmd)


def install(source, package_id):
    """
    Install a .pkg from an URI or an absolute path.

    :param str source: The path to a package.

    :param str package_id: The package ID

    :return: True if successful, otherwise False
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' pkgutil.install source=/vagrant/build_essentials.pkg package_id=com.apple.pkg.gcc4.2Leo
    """
    if is_installed(package_id):
        return True

    uri = urllib.parse.urlparse(source)
    if not uri.scheme == "":
        msg = f"Unsupported scheme for source uri: {uri.scheme}"
        raise SaltInvocationError(msg)

    _install_from_path(source)

    return is_installed(package_id)


def forget(package_id):
    """
    .. versionadded:: 2016.3.0

    Remove the receipt data about the specified package. Does not remove files.

    .. warning::
        DO NOT use this command to fix broken package design

    :param str package_id: The name of the package to forget

    :return: True if successful, otherwise False
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' pkgutil.forget com.apple.pkg.gcc4.2Leo
    """
    cmd = f"pkgutil --forget {package_id}"
    salt.utils.mac_utils.execute_return_success(cmd)
    return not is_installed(package_id)



import textwrap

import pytest

import salt.utils.platform
from tests.support.runtests import RUNTIME_VARS

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def minion_id():
    return "terraform_ssh_minion"


@pytest.fixture(scope="module")
def terraform_roster_file(sshd_server, salt_master, tmp_path_factory, minion_id):
    darwin_addon = ""
    if salt.utils.platform.is_darwin():
        darwin_addon = ',\n        "set_path": "$PATH:/usr/local/bin/"\n'
    roster_contents = textwrap.dedent(
        """    {{
      "version": 4,
      "terraform_version": "1.4.3",
      "serial": 1,
      "outputs": {{}},
      "resources": [
        {{
          "mode": "managed",
          "type": "salt_host",
          "name": "{minion}",
          "instances": [
            {{
              "schema_version": 0,
              "attributes": {{
                "cmd_umask": null,
                "host": "localhost",
                "id": "{minion}",
                "minion_opts": null,
                "passwd": "",
                "port": {port},
                "priv": null,
                "salt_id": "{minion}",
                "sudo": null,
                "sudo_user": null,
                "thin_dir": null,
                "timeout": null,
                "tty": null,
                "user": "{user}"{darwin_addon}
              }}
            }}
          ]
        }}
      ],
      "check_results": null
    }}
    """
    ).format(
        minion=minion_id,
        port=sshd_server.listen_port,
        user=RUNTIME_VARS.RUNNING_TESTS_USER,
        darwin_addon=darwin_addon,
    )
    roster_file = tmp_path_factory.mktemp("terraform_roster") / "terraform.tfstate"
    roster_file.write_text(roster_contents)
    yield roster_file
    roster_file.unlink()


@pytest.fixture(scope="module")
def salt_ssh_cli(salt_master, terraform_roster_file, sshd_config_dir):
    """
    The ``salt-ssh`` CLI as a fixture against the running master
    """
    assert salt_master.is_running()
    return salt_master.salt_ssh_cli(
        roster_file=terraform_roster_file,
        target_host="*",
        client_key=str(sshd_config_dir / "client_key"),
        base_script_args=["--ignore-host-keys"],
    )


def test_terraform_roster(salt_ssh_cli, minion_id):
    """
    Test that the terraform roster operates as intended
    """
    ret = salt_ssh_cli.run("--roster=terraform", "test.ping")
    assert ret.data.get(minion_id) is True



"""
Support for Open vSwitch - module with basic Open vSwitch commands.

Suitable for setting up Openstack Neutron.

:codeauthor: Jiri Kotlin <jiri.kotlin@ultimum.io>
"""

import logging

import salt.utils.path
from salt.exceptions import ArgumentValueError, CommandExecutionError
from salt.utils import json

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load the module if Open vSwitch is installed
    """
    if salt.utils.path.which("ovs-vsctl"):
        return "openvswitch"
    return (False, "Missing dependency: ovs-vsctl")


def _param_may_exist(may_exist):
    """
    Returns --may-exist parameter for Open vSwitch command.

    Args:
        may_exist: Boolean whether to use this parameter.

    Returns:
        String '--may-exist ' or empty string.
    """
    if may_exist:
        return "--may-exist "
    else:
        return ""


def _param_if_exists(if_exists):
    """
    Returns --if-exist parameter for Open vSwitch command.

    Args:
        if_exists: Boolean whether to use this parameter.

    Returns:
        String '--if-exist ' or empty string.
    """
    if if_exists:
        return "--if-exists "
    else:
        return ""


def _retcode_to_bool(retcode):
    """
    Evaulates Open vSwitch command`s retcode value.

    Args:
        retcode: Value of retcode field from response, should be 0, 1 or 2.

    Returns:
        True on 0, else False
    """
    if retcode == 0:
        return True
    else:
        return False


def _stdout_list_split(retcode, stdout="", splitstring="\n"):
    """
    Evaulates Open vSwitch command`s retcode value.

    Args:
        retcode: Value of retcode field from response, should be 0, 1 or 2.
        stdout: Value of stdout filed from response.
        splitstring: String used to split the stdout default new line.

    Returns:
        List or False.
    """
    if retcode == 0:
        ret = stdout.split(splitstring)
        return ret
    else:
        return False


def _convert_json(obj):
    """
    Converts from the JSON output provided by ovs-vsctl into a usable Python
    object tree. In particular, sets and maps are converted from lists to
    actual sets or maps.

    Args:
        obj: Object that shall be recursively converted.

    Returns:
        Converted version of object.
    """
    if isinstance(obj, dict):
        return {_convert_json(key): _convert_json(val) for (key, val) in obj.items()}
    elif isinstance(obj, list) and len(obj) == 2:
        first = obj[0]
        second = obj[1]
        if first == "set" and isinstance(second, list):
            return [_convert_json(elem) for elem in second]
        elif first == "map" and isinstance(second, list):
            for elem in second:
                if not isinstance(elem, list) or len(elem) != 2:
                    return obj
            return {elem[0]: _convert_json(elem[1]) for elem in second}
        else:
            return obj
    elif isinstance(obj, list):
        return [_convert_json(elem) for elem in obj]
    else:
        return obj


def _stdout_parse_json(stdout):
    """
    Parses JSON output from ovs-vsctl and returns the corresponding object
    tree.

    Args:
        stdout: Output that shall be parsed.

    Returns:
        Object represented by the output.
    """
    obj = json.loads(stdout)
    return _convert_json(obj)


def bridge_list():
    """
    Lists all existing real and fake bridges.

    Returns:
        List of bridges (or empty list), False on failure.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.bridge_list
    """
    cmd = "ovs-vsctl list-br"
    result = __salt__["cmd.run_all"](cmd)
    retcode = result["retcode"]
    stdout = result["stdout"]
    return _stdout_list_split(retcode, stdout)


def bridge_exists(br):
    """
    Tests whether bridge exists as a real or fake  bridge.

    Returns:
        True if Bridge exists, else False.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.bridge_exists br0
    """
    cmd = f"ovs-vsctl br-exists {br}"
    result = __salt__["cmd.run_all"](cmd)
    retcode = result["retcode"]
    return _retcode_to_bool(retcode)


def bridge_create(br, may_exist=True, parent=None, vlan=None):
    """
    Creates a new bridge.

    Args:
        br : string
            bridge name
        may_exist : bool
            if False - attempting to create a bridge that exists returns False.
        parent : string
            name of the parent bridge (if the bridge shall be created as a fake
            bridge). If specified, vlan must also be specified.
        .. versionadded:: 3006.0
        vlan : int
            VLAN ID of the bridge (if the bridge shall be created as a fake
            bridge). If specified, parent must also be specified.
        .. versionadded:: 3006.0

    Returns:
        True on success, else False.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.bridge_create br0
    """
    param_may_exist = _param_may_exist(may_exist)
    if parent is not None and vlan is None:
        raise ArgumentValueError("If parent is specified, vlan must also be specified.")
    if vlan is not None and parent is None:
        raise ArgumentValueError("If vlan is specified, parent must also be specified.")
    param_parent = "" if parent is None else f" {parent}"
    param_vlan = "" if vlan is None else f" {vlan}"
    cmd = "ovs-vsctl {1}add-br {0}{2}{3}".format(
        br, param_may_exist, param_parent, param_vlan
    )
    result = __salt__["cmd.run_all"](cmd)
    return _retcode_to_bool(result["retcode"])


def bridge_delete(br, if_exists=True):
    """
    Deletes bridge and all of  its  ports.

    Args:
        br: A string - bridge name
        if_exists: Bool, if False - attempting to delete a bridge that does not exist returns False.

    Returns:
        True on success, else False.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.bridge_delete br0
    """
    param_if_exists = _param_if_exists(if_exists)
    cmd = f"ovs-vsctl {param_if_exists}del-br {br}"
    result = __salt__["cmd.run_all"](cmd)
    retcode = result["retcode"]
    return _retcode_to_bool(retcode)


def bridge_to_parent(br):
    """
    .. versionadded:: 3006.0

    Returns the parent bridge of a bridge.

    Args:
        br : string
            bridge name

    Returns:
        Name of the parent bridge. This is the same as the bridge name if the
        bridge is not a fake bridge. If the bridge does not exist, False is
        returned.

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.bridge_to_parent br0
    """
    cmd = f"ovs-vsctl br-to-parent {br}"
    result = __salt__["cmd.run_all"](cmd)
    if result["retcode"] != 0:
        return False
    return result["stdout"].strip()


def bridge_to_vlan(br):
    """
    .. versionadded:: 3006.0

    Returns the VLAN ID of a bridge.

    Args:
        br : string
            bridge name

    Returns:
        VLAN ID of the bridge. The VLAN ID is 0 if the bridge is not a fake
        bridge.  If the bridge does not exist, False is returned.

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.bridge_to_parent br0
    """
    cmd = f"ovs-vsctl br-to-vlan {br}"
    result = __salt__["cmd.run_all"](cmd)
    if result["retcode"] != 0:
        return False
    return int(result["stdout"])


def port_add(br, port, may_exist=False, internal=False):
    """
    Creates on bridge a new port named port.

    Returns:
        True on success, else False.

    Args:
        br: A string - bridge name
        port: A string - port name
        may_exist: Bool, if False - attempting to create a port that exists returns False.
        internal: A boolean to create an internal interface if one does not exist.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.port_add br0 8080
    """
    param_may_exist = _param_may_exist(may_exist)
    cmd = f"ovs-vsctl {param_may_exist}add-port {br} {port}"
    if internal:
        cmd += f" -- set interface {port} type=internal"
    result = __salt__["cmd.run_all"](cmd)
    retcode = result["retcode"]
    return _retcode_to_bool(retcode)


def port_remove(br, port, if_exists=True):
    """
     Deletes port.

    Args:
        br: A string - bridge name (If bridge is None, port is removed from  whatever bridge contains it)
        port: A string - port name.
        if_exists: Bool, if False - attempting to delete a por that  does  not exist returns False. (Default True)

    Returns:
        True on success, else False.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.port_remove br0 8080
    """
    param_if_exists = _param_if_exists(if_exists)

    if port and not br:
        cmd = f"ovs-vsctl {param_if_exists}del-port {port}"
    else:
        cmd = f"ovs-vsctl {param_if_exists}del-port {br} {port}"
    result = __salt__["cmd.run_all"](cmd)
    retcode = result["retcode"]
    return _retcode_to_bool(retcode)


def port_list(br):
    """
    Lists all of the ports within bridge.

    Args:
        br: A string - bridge name.

    Returns:
        List of bridges (or empty list), False on failure.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.port_list br0
    """
    cmd = f"ovs-vsctl list-ports {br}"
    result = __salt__["cmd.run_all"](cmd)
    retcode = result["retcode"]
    stdout = result["stdout"]
    return _stdout_list_split(retcode, stdout)


def port_get_tag(port):
    """
    Lists tags of the port.

    Args:
        port: A string - port name.

    Returns:
        List of tags (or empty list), False on failure.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.port_get_tag tap0
    """
    cmd = f"ovs-vsctl get port {port} tag"
    result = __salt__["cmd.run_all"](cmd)
    retcode = result["retcode"]
    stdout = result["stdout"]
    return _stdout_list_split(retcode, stdout)


def interface_get_options(port):
    """
    Port's interface's optional parameters.

    Args:
        port: A string - port name.

    Returns:
        String containing optional parameters of port's interface, False on failure.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.interface_get_options tap0
    """
    cmd = f"ovs-vsctl get interface {port} options"
    result = __salt__["cmd.run_all"](cmd)
    retcode = result["retcode"]
    stdout = result["stdout"]
    return _stdout_list_split(retcode, stdout)


def interface_get_type(port):
    """
    Type of port's interface.

    Args:
        port: A string - port name.

    Returns:
        String - type of interface or empty string, False on failure.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' openvswitch.interface_get_type tap0
    """
    cmd = f"ovs-vsctl get interface {port} type"
    result = __salt__["cmd.run_all"](cmd)
    retcode = result["retcode"]
    stdout = result["stdout"]
    return _stdout_list_split(retcode, stdout)


def port_create_vlan(br, port, id, internal=False):
    """
    Isolate VM traffic using VLANs.

    Args:
        br: A string - bridge name.
        port: A string - port name.
        id: An integer in the valid range 0 to 4095 (inclusive), name of VLAN.
        internal: A boolean to create an internal interface if one does not exist.

    Returns:
        True on success, else False.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

       salt '*' openvswitch.port_create_vlan br0 tap0 100
    """
    interfaces = __salt__["network.interfaces"]()
    if not 0 <= id <= 4095:
        return False
    elif not bridge_exists(br):
        return False
    elif not internal and port not in interfaces:
        return False
    elif port in port_list(br):
        cmd = f"ovs-vsctl set port {port} tag={id}"
        if internal:
            cmd += f" -- set interface {port} type=internal"
        result = __salt__["cmd.run_all"](cmd)
        return _retcode_to_bool(result["retcode"])
    else:
        cmd = f"ovs-vsctl add-port {br} {port} tag={id}"
        if internal:
            cmd += f" -- set interface {port} type=internal"
        result = __salt__["cmd.run_all"](cmd)
        return _retcode_to_bool(result["retcode"])


def port_create_gre(br, port, id, remote):
    """
    Generic Routing Encapsulation - creates GRE tunnel between endpoints.

    Args:
        br: A string - bridge name.
        port: A string - port name.
        id: An integer - unsigned 32-bit number, tunnel's key.
        remote: A string - remote endpoint's IP address.

    Returns:
        True on success, else False.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

       salt '*' openvswitch.port_create_gre br0 gre1 5001 192.168.1.10
    """
    if not 0 <= id < 2**32:
        return False
    elif not __salt__["dig.check_ip"](remote):
        return False
    elif not bridge_exists(br):
        return False
    elif port in port_list(br):
        cmd = "ovs-vsctl set interface {} type=gre options:remote_ip={} options:key={}".format(
            port, remote, id
        )
        result = __salt__["cmd.run_all"](cmd)
        return _retcode_to_bool(result["retcode"])
    else:
        cmd = (
            "ovs-vsctl add-port {0} {1} -- set interface {1} type=gre"
            " options:remote_ip={2} options:key={3}".format(br, port, remote, id)
        )
        result = __salt__["cmd.run_all"](cmd)
        return _retcode_to_bool(result["retcode"])


def port_create_vxlan(br, port, id, remote, dst_port=None):
    """
    Virtual eXtensible Local Area Network - creates VXLAN tunnel between endpoints.

    Args:
        br: A string - bridge name.
        port: A string - port name.
        id: An integer - unsigned 64-bit number, tunnel's key.
        remote: A string - remote endpoint's IP address.
        dst_port: An integer - port to use when creating tunnelport in the switch.

    Returns:
        True on success, else False.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

       salt '*' openvswitch.port_create_vxlan br0 vx1 5001 192.168.1.10 8472
    """
    dst_port = " options:dst_port=" + str(dst_port) if 0 < dst_port <= 65535 else ""
    if not 0 <= id < 2**64:
        return False
    elif not __salt__["dig.check_ip"](remote):
        return False
    elif not bridge_exists(br):
        return False
    elif port in port_list(br):
        cmd = (
            "ovs-vsctl set interface {} type=vxlan options:remote_ip={} "
            "options:key={}{}".format(port, remote, id, dst_port)
        )
        result = __salt__["cmd.run_all"](cmd)
        return _retcode_to_bool(result["retcode"])
    else:
        cmd = (
            "ovs-vsctl add-port {0} {1} -- set interface {1} type=vxlan"
            " options:remote_ip={2} options:key={3}{4}".format(
                br, port, remote, id, dst_port
            )
        )
        result = __salt__["cmd.run_all"](cmd)
        return _retcode_to_bool(result["retcode"])


def db_get(table, record, column, if_exists=False):
    """
    .. versionadded:: 3006.0

    Gets a column's value for a specific record.

    Args:
        table : string
            name of the database table
        record : string
            identifier of the record
        column : string
            name of the column
        if_exists : boolean
            if True, it is not an error if the record does not exist.

    Returns:
        The column's value.

    CLI Example:

    .. code-block:: bash

       salt '*' openvswitch.db_get Port br0 vlan_mode
    """
    cmd = ["ovs-vsctl", "--format=json", f"--columns={column}"]
    if if_exists:
        cmd += ["--if-exists"]
    cmd += ["list", table, record]
    result = __salt__["cmd.run_all"](cmd)
    if result["retcode"] != 0:
        raise CommandExecutionError(result["stderr"])
    output = _stdout_parse_json(result["stdout"])
    if output["data"] and output["data"][0]:
        return output["data"][0][0]
    else:
        return None


def db_set(table, record, column, value, if_exists=False):
    """
    .. versionadded:: 3006.0

    Sets a column's value for a specific record.

    Args:
        table : string
            name of the database table
        record : string
            identifier of the record
        column : string
            name of the column
        value : string
            the value to be set
        if_exists : boolean
            if True, it is not an error if the record does not exist.

    Returns:
        None on success and an error message on failure.

    CLI Example:

    .. code-block:: bash

       salt '*' openvswitch.db_set Interface br0 mac 02:03:04:05:06:07
    """
    cmd = ["ovs-vsctl"]
    if if_exists:
        cmd += ["--if-exists"]
    cmd += ["set", table, record, f"{column}={json.dumps(value)}"]
    result = __salt__["cmd.run_all"](cmd)
    if result["retcode"] != 0:
        return result["stderr"]
    else:
        return None



"""
Send events based on a script's stdout

Example Config

.. code-block:: yaml

    engines:
      - script:
          cmd: /some/script.py -a 1 -b 2
          output: json
          interval: 5
          onchange: false

Script engine configs:

cmd
    Script or command to execute

output
    Any available saltstack deserializer

interval
    How often in seconds to execute the command

onchange
    .. versionadded:: 3006.0

    Only fire an event if the tag-specific output changes. Defaults to False.
"""

import logging
import shlex
import subprocess
import time

import salt.loader
import salt.utils.event
import salt.utils.process
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def _read_stdout(proc):
    """
    Generator that returns stdout
    """
    yield from iter(proc.stdout.readline, b"")


def _get_serializer(output):
    """
    Helper to return known serializer based on
    pass output argument
    """
    serializers = salt.loader.serializers(__opts__)
    try:
        return getattr(serializers, output)
    except AttributeError:
        raise CommandExecutionError(
            f"Unknown serializer `{output}` found for output option"
        )


def start(cmd, output="json", interval=1, onchange=False):
    """
    Parse stdout of a command and generate an event

    The script engine will scrap stdout of the
    given script and generate an event based on the
    presence of the 'tag' key and its value.

    If there is a data obj available, that will also
    be fired along with the tag.

    Example:

        Given the following json output from a script:

            .. code-block:: json

                { "tag" : "lots/of/tacos",
                "data" : { "toppings" : "cilantro" }
                }

        This will fire the event 'lots/of/tacos'
        on the event bus with the data obj as is.

    :param cmd: The command to execute
    :param output: How to deserialize stdout of the script
    :param interval: How often to execute the script
    :param onchange: Only fire an event if the tag-specific output changes
    """
    try:
        cmd = shlex.split(cmd)
    except AttributeError:
        cmd = shlex.split(str(cmd))
    log.debug("script engine using command %s", cmd)

    serializer = _get_serializer(output)

    if __opts__.get("__role") == "master":
        fire_master = salt.utils.event.get_master_event(
            __opts__, __opts__["sock_dir"]
        ).fire_event
    else:
        fire_master = __salt__["event.send"]

    if onchange:
        events = {}

    while True:
        try:
            proc = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )

            log.debug("Starting script with pid %d", proc.pid)

            for raw_event in _read_stdout(proc):
                log.debug(raw_event)

                event = serializer.deserialize(raw_event)
                tag = event.get("tag", None)
                data = event.get("data", {})

                if data and "id" not in data:
                    data["id"] = __opts__["id"]

                if tag:
                    if onchange and tag in events and events[tag] == data:
                        continue
                    log.info("script engine firing event with tag %s", tag)
                    fire_master(tag=tag, data=data)
                    if onchange:
                        events[tag] = data

            log.debug("Closing script with pid %d", proc.pid)
            proc.stdout.close()
            rc = proc.wait()
            if rc:
                raise subprocess.CalledProcessError(rc, cmd)

        except subprocess.CalledProcessError as e:
            log.error(e)
        finally:
            if proc.poll is None:
                proc.terminate()

        time.sleep(interval)



"""
Module for managing the LXD daemon and its containers.

.. versionadded:: 2019.2.0

`LXD(1)`_ is a container "hypervisor". This execution module provides
several functions to help manage it and its containers.

.. note::

    - `pylxd(2)`_ version >=2.2.5 is required to let this work,
      currently only available via pip.

        To install on Ubuntu:

        $ apt-get install libssl-dev python-pip
        $ pip install -U pylxd

    - you need lxd installed on the minion
      for the init() and version() methods.

    - for the config_get() and config_get() methods
      you need to have lxd-client installed.

.. _LXD(1): https://linuxcontainers.org/lxd/
.. _pylxd(2): https://github.com/lxc/pylxd/blob/master/doc/source/installation.rst

:maintainer: René Jochum <rene@jochums.at>
:maturity: new
:depends: python-pylxd
:platform: Linux
"""

import logging
import os
from datetime import datetime

import salt.utils.decorators.path
import salt.utils.files
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.utils.versions import Version

try:
    import pylxd

    HAS_PYLXD = True

    import urllib3

    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
except ImportError:
    HAS_PYLXD = False


log = logging.getLogger(__name__)

__docformat__ = "restructuredtext en"

_pylxd_minimal_version = "2.2.5"

# Keep in sync with: https://github.com/lxc/lxd/blob/master/shared/osarch/architectures.go
_architectures = {
    "unknown": "0",
    "i686": "1",
    "x86_64": "2",
    "armv7l": "3",
    "aarch64": "4",
    "ppc": "5",
    "ppc64": "6",
    "ppc64le": "7",
    "s390x": "8",
}

# Keep in sync with: https://github.com/lxc/lxd/blob/master/shared/api/status_code.go
CONTAINER_STATUS_RUNNING = 103

__virtualname__ = "lxd"

_connection_pool = {}


def __virtual__():
    if HAS_PYLXD:
        if Version(pylxd_version()) < Version(_pylxd_minimal_version):
            return (
                False,
                'The lxd execution module cannot be loaded: pylxd "{}" is '
                'not supported, you need at least pylxd "{}"'.format(
                    pylxd_version(), _pylxd_minimal_version
                ),
            )

        return __virtualname__

    return (
        False,
        "The lxd execution module cannot be loaded: "
        "the pylxd python module is not available.",
    )


################
# LXD Management
################
@salt.utils.decorators.path.which("lxd")
def version():
    """
    Returns the actual lxd version.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.version

    """
    return __salt__["cmd.run"]("lxd --version")


def pylxd_version():
    """
    Returns the actual pylxd version.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.pylxd_version

    """
    return pylxd.__version__


@salt.utils.decorators.path.which("lxd")
def init(
    storage_backend="dir",
    trust_password=None,
    network_address=None,
    network_port=None,
    storage_create_device=None,
    storage_create_loop=None,
    storage_pool=None,
):
    """
    Calls lxd init --auto -- opts

    storage_backend :
        Storage backend to use (zfs or dir, default: dir)

    trust_password :
        Password required to add new clients

    network_address : None
        Address to bind LXD to (default: none)

    network_port : None
        Port to bind LXD to (Default: 8443)

    storage_create_device : None
        Setup device based storage using this DEVICE

    storage_create_loop : None
        Setup loop based storage with this SIZE in GB

    storage_pool : None
        Storage pool to use or create

    CLI Examples:

    To listen on all IPv4/IPv6 Addresses:

    .. code-block:: bash

        salt '*' lxd.init dir PaSsW0rD [::]

    To not listen on Network:

    .. code-block:: bash

        salt '*' lxd.init
    """

    cmd = f'lxd init --auto --storage-backend="{storage_backend}"'

    if trust_password is not None:
        cmd = cmd + f' --trust-password="{trust_password}"'

    if network_address is not None:
        cmd = cmd + f' --network-address="{network_address}"'

    if network_port is not None:
        cmd = cmd + f' --network-port="{network_port}"'

    if storage_create_device is not None:
        cmd = cmd + f' --storage-create-device="{storage_create_device}"'

    if storage_create_loop is not None:
        cmd = cmd + f' --storage-create-loop="{storage_create_loop}"'

    if storage_pool is not None:
        cmd = cmd + f' --storage-pool="{storage_pool}"'

    try:
        output = __salt__["cmd.run"](cmd)
    except ValueError as e:
        raise CommandExecutionError(
            f"Failed to call: '{cmd}', error was: {str(e)}",
        )

    if "error:" in output:
        raise CommandExecutionError(
            output[output.index("error:") + 7 :],
        )

    return output


@salt.utils.decorators.path.which("lxd")
@salt.utils.decorators.path.which("lxc")
def config_set(key, value):
    """
    Set an LXD daemon config option

    CLI Examples:

    To listen on IPv4 and IPv6 port 8443,
    you can omit the :8443 its the default:

    .. code-block:: bash

        salt '*' lxd.config_set core.https_address [::]:8443

    To set the server trust password:

    .. code-block:: bash

        salt '*' lxd.config_set core.trust_password blah

    """
    cmd = 'lxc config set "{}" "{}"'.format(
        key,
        value,
    )

    output = __salt__["cmd.run"](cmd)
    if "error:" in output:
        raise CommandExecutionError(
            output[output.index("error:") + 7 :],
        )

    return (f'Config value "{key}" successfully set.',)


@salt.utils.decorators.path.which("lxd")
@salt.utils.decorators.path.which("lxc")
def config_get(key):
    """
    Get an LXD daemon config option

    key :
        The key of the config value to retrieve

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.config_get core.https_address
    """

    cmd = f'lxc config get "{key}"'

    output = __salt__["cmd.run"](cmd)
    if "error:" in output:
        raise CommandExecutionError(
            output[output.index("error:") + 7 :],
        )

    return output


#######################
# Connection Management
#######################
def pylxd_client_get(remote_addr=None, cert=None, key=None, verify_cert=True):
    """
    Get an pyxld client, this is not meant to be run over the CLI.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if you
        provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    See the `requests-docs`_ for the SSL stuff.

    .. _requests-docs: http://docs.python-requests.org/en/master/user/advanced/#ssl-cert-verification

    """

    pool_key = "|".join(
        (
            str(remote_addr),
            str(cert),
            str(key),
            str(verify_cert),
        )
    )

    if pool_key in _connection_pool:
        log.debug('Returning the client "%s" from our connection pool', remote_addr)
        return _connection_pool[pool_key]

    try:
        if remote_addr is None or remote_addr == "/var/lib/lxd/unix.socket":
            log.debug("Trying to connect to the local unix socket")
            client = pylxd.Client()
        else:
            if remote_addr.startswith("/"):
                client = pylxd.Client(remote_addr)
            else:
                if cert is None or key is None:
                    raise SaltInvocationError(
                        "You have to give a Cert and Key file for remote endpoints."
                    )

                cert = os.path.expanduser(cert)
                key = os.path.expanduser(key)

                if not os.path.isfile(cert):
                    raise SaltInvocationError(
                        'You have given an invalid cert path: "{}", the '
                        "file does not exist or is not a file.".format(cert)
                    )

                if not os.path.isfile(key):
                    raise SaltInvocationError(
                        'You have given an invalid key path: "{}", the '
                        "file does not exists or is not a file.".format(key)
                    )

                log.debug(
                    'Trying to connect to "%s" with cert "%s", key "%s" and '
                    'verify_cert "%s"',
                    remote_addr,
                    cert,
                    key,
                    verify_cert,
                )
                client = pylxd.Client(
                    endpoint=remote_addr,
                    cert=(
                        cert,
                        key,
                    ),
                    verify=verify_cert,
                )
    except pylxd.exceptions.ClientConnectionFailed:
        raise CommandExecutionError(f"Failed to connect to '{remote_addr}'")

    except TypeError as e:
        # Happens when the verification failed.
        raise CommandExecutionError(
            'Failed to connect to "{}", looks like the SSL verification '
            "failed, error was: {}".format(remote_addr, str(e))
        )

    _connection_pool[pool_key] = client

    return client


def pylxd_save_object(obj):
    """Saves an object (profile/image/container) and
        translate its execpetion on failure

    obj :
        The object to save

    This is an internal method, no CLI Example.
    """
    try:
        obj.save(wait=True)
    except pylxd.exceptions.LXDAPIException as e:
        raise CommandExecutionError(str(e))

    return True


def authenticate(remote_addr, password, cert, key, verify_cert=True):
    """
    Authenticate with a remote LXDaemon.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if you
        provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443

    password :
        The password of the remote.

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.authenticate https://srv01:8443 <yourpass> ~/.config/lxc/client.crt ~/.config/lxc/client.key false

    See the `requests-docs`_ for the SSL stuff.

    .. _requests-docs: http://docs.python-requests.org/en/master/user/advanced/#ssl-cert-verification

    """
    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    if client.trusted:
        return True

    try:
        client.authenticate(password)
    except pylxd.exceptions.LXDAPIException as e:
        # Wrong password
        raise CommandExecutionError(str(e))

    return client.trusted


######################
# Container Management
######################
def container_list(
    list_names=False, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """
    Lists containers

    list_names : False
        Only return a list of names when True

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Examples:

    Full dict with all available information:

    .. code-block:: bash

        salt '*' lxd.container_list

    For a list of names:

    .. code-block:: bash

        salt '*' lxd.container_list true

    See also `container-attributes`_.

    .. _container-attributes: https://github.com/lxc/pylxd/blob/master/doc/source/containers.rst#container-attributes

    """

    client = pylxd_client_get(remote_addr, cert, key, verify_cert)
    containers = client.containers.all()
    if list_names:
        return [c.name for c in containers]

    return map(_pylxd_model_to_dict, containers)


def container_create(
    name,
    source,
    profiles=None,
    config=None,
    devices=None,
    architecture="x86_64",
    ephemeral=False,
    wait=True,
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
    _raw=False,
):
    """
    Create a container

    name :
        The name of the container

    source :
        Can be either a string containing an image alias:
             "xenial/amd64"

        or an dict with type "image" with alias:
            {"type": "image",
             "alias": "xenial/amd64"}

        or image with "fingerprint":
            {"type": "image",
             "fingerprint": "SHA-256"}

        or image with "properties":
            {"type": "image",
             "properties": {
                "os": "ubuntu",
                "release": "14.04",
                "architecture": "x86_64"}}

        or none:
            {"type": "none"}

        or copy:
            {"type": "copy",
             "source": "my-old-container"}

    profiles : ['default']
        List of profiles to apply on this container

    config :
        A config dict or None (None = unset).

        Can also be a list:
            [{'key': 'boot.autostart', 'value': 1},
             {'key': 'security.privileged', 'value': '1'}]

    devices :
        A device dict or None (None = unset).

    architecture : 'x86_64'
        Can be one of the following:
            * unknown
            * i686
            * x86_64
            * armv7l
            * aarch64
            * ppc
            * ppc64
            * ppc64le
            * s390x

    ephemeral : False
        Destroy this container after stop?

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    _raw : False
        Return the raw pyxld object or a dict?

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.container_create test xenial/amd64

    See also the `rest-api-docs`_.

    .. _rest-api-docs: https://github.com/lxc/lxd/blob/master/doc/rest-api.md#post-1

    """
    if profiles is None:
        profiles = ["default"]

    if config is None:
        config = {}

    if devices is None:
        devices = {}

    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    if not isinstance(profiles, (list, tuple, set)):
        raise SaltInvocationError("'profiles' must be formatted as list/tuple/set.")

    if architecture not in _architectures:
        raise SaltInvocationError(
            "Unknown architecture '{}' given for the container '{}'".format(
                architecture, name
            )
        )

    if isinstance(source, str):
        source = {"type": "image", "alias": source}

    config, devices = normalize_input_values(config, devices)

    try:
        container = client.containers.create(
            {
                "name": name,
                "architecture": _architectures[architecture],
                "profiles": profiles,
                "source": source,
                "config": config,
                "ephemeral": ephemeral,
            },
            wait=wait,
        )
    except pylxd.exceptions.LXDAPIException as e:
        raise CommandExecutionError(str(e))

    if not wait:
        return container.json()["operation"]

    # Add devices if not wait and devices have been given.
    if devices:
        for dn, dargs in devices.items():
            container_device_add(name, dn, **dargs)

    if _raw:
        return container

    return _pylxd_model_to_dict(container)


def container_get(
    name=None, remote_addr=None, cert=None, key=None, verify_cert=True, _raw=False
):
    """Gets a container from the LXD

    name :
        The name of the container to get.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    _raw :
        Return the pylxd object, this is internal and by states in use.
    """
    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    if name is None:
        containers = client.containers.all()
        if _raw:
            return containers
    else:
        containers = []
        try:
            containers = [client.containers.get(name)]
        except pylxd.exceptions.LXDAPIException:
            raise SaltInvocationError(f"Container '{name}' not found")
        if _raw:
            return containers[0]

    infos = []
    for container in containers:
        infos.append(dict([(container.name, _pylxd_model_to_dict(container))]))
    return infos


def container_delete(name, remote_addr=None, cert=None, key=None, verify_cert=True):
    """
    Delete a container

    name :
        Name of the container to delete

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)
    container.delete(wait=True)
    return True


def container_rename(
    name, newname, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """
    Rename a container

    name :
        Name of the container to Rename

    newname :
        The new name of the container

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    if container.status_code == CONTAINER_STATUS_RUNNING:
        raise SaltInvocationError(f"Can't rename the running container '{name}'.")

    container.rename(newname, wait=True)
    return _pylxd_model_to_dict(container)


def container_state(name=None, remote_addr=None, cert=None, key=None, verify_cert=True):
    """
    Get container state

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    if name is None:
        containers = client.containers.all()
    else:
        try:
            containers = [client.containers.get(name)]
        except pylxd.exceptions.LXDAPIException:
            raise SaltInvocationError(f"Container '{name}' not found")

    states = []
    for container in containers:
        state = {}
        state = container.state()

        states.append(
            dict(
                [
                    (
                        container.name,
                        {
                            k: getattr(state, k)
                            for k in dir(state)
                            if not k.startswith("_")
                        },
                    )
                ]
            )
        )
    return states


def container_start(name, remote_addr=None, cert=None, key=None, verify_cert=True):
    """
    Start a container

    name :
        Name of the container to start

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)
    container.start(wait=True)
    return _pylxd_model_to_dict(container)


def container_stop(
    name,
    timeout=30,
    force=True,
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
):
    """
    Stop a container

    name :
        Name of the container to stop

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)
    container.stop(timeout, force, wait=True)
    return _pylxd_model_to_dict(container)


def container_restart(name, remote_addr=None, cert=None, key=None, verify_cert=True):
    """
    Restart a container

    name :
        Name of the container to restart

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)
    container.restart(wait=True)
    return _pylxd_model_to_dict(container)


def container_freeze(name, remote_addr=None, cert=None, key=None, verify_cert=True):
    """
    Freeze a container

    name :
        Name of the container to freeze

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)
    container.freeze(wait=True)
    return _pylxd_model_to_dict(container)


def container_unfreeze(name, remote_addr=None, cert=None, key=None, verify_cert=True):
    """
    Unfreeze a container

    name :
        Name of the container to unfreeze

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)
    container.unfreeze(wait=True)
    return _pylxd_model_to_dict(container)


def container_migrate(
    name,
    stop_and_start=False,
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
    src_remote_addr=None,
    src_cert=None,
    src_key=None,
    src_verify_cert=None,
):
    """Migrate a container.

    If the container is running, it either must be shut down
    first (use stop_and_start=True) or criu must be installed
    on the source and destination machines.

    For this operation both certs need to be authenticated,
    use :mod:`lxd.authenticate <salt.modules.lxd.authenticate`
    to authenticate your cert(s).

    name :
        Name of the container to migrate

    stop_and_start :
        Stop the container on the source and start it on dest

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Example:

    .. code-block:: bash

        # Authorize
        salt '*' lxd.authenticate https://srv01:8443 <yourpass> ~/.config/lxc/client.crt ~/.config/lxc/client.key false
        salt '*' lxd.authenticate https://srv02:8443 <yourpass> ~/.config/lxc/client.crt ~/.config/lxc/client.key false

        # Migrate phpmyadmin from srv01 to srv02
        salt '*' lxd.container_migrate phpmyadmin stop_and_start=true remote_addr=https://srv02:8443 cert=~/.config/lxc/client.crt key=~/.config/lxc/client.key verify_cert=False src_remote_addr=https://srv01:8443
    """
    if src_cert is None:
        src_cert = cert

    if src_key is None:
        src_key = key

    if src_verify_cert is None:
        src_verify_cert = verify_cert

    container = container_get(
        name, src_remote_addr, src_cert, src_key, src_verify_cert, _raw=True
    )

    dest_client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    for pname in container.profiles:
        try:
            dest_client.profiles.get(pname)
        except pylxd.exceptions.LXDAPIException:
            raise SaltInvocationError(
                "not all the profiles from the source exist on the target"
            )

    was_running = container.status_code == CONTAINER_STATUS_RUNNING
    if stop_and_start and was_running:
        container.stop(wait=True)

    try:
        dest_container = container.migrate(dest_client, wait=True)
        dest_container.profiles = container.profiles
        dest_container.save()
    except pylxd.exceptions.LXDAPIException as e:
        raise CommandExecutionError(str(e))

    # Remove the source container
    container.delete(wait=True)

    if stop_and_start and was_running:
        dest_container.start(wait=True)

    return _pylxd_model_to_dict(dest_container)


def container_config_get(
    name, config_key, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """
    Get a container config value

    name :
        Name of the container

    config_key :
        The config key to retrieve

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)
    return _get_property_dict_item(container, "config", config_key)


def container_config_set(
    name,
    config_key,
    config_value,
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
):
    """
    Set a container config value

    name :
        Name of the container

    config_key :
        The config key to set

    config_value :
        The config value to set

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    return _set_property_dict_item(container, "config", config_key, config_value)


def container_config_delete(
    name, config_key, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """
    Delete a container config value

    name :
        Name of the container

    config_key :
        The config key to delete

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    return _delete_property_dict_item(container, "config", config_key)


def container_device_get(
    name, device_name, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """
    Get a container device

    name :
        Name of the container

    device_name :
        The device name to retrieve

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    return _get_property_dict_item(container, "devices", device_name)


def container_device_add(
    name,
    device_name,
    device_type="disk",
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
    **kwargs,
):
    """
    Add a container device

    name :
        Name of the container

    device_name :
        The device name to add

    device_type :
        Type of the device

    ** kwargs :
        Additional device args

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    kwargs["type"] = device_type
    return _set_property_dict_item(container, "devices", device_name, kwargs)


def container_device_delete(
    name, device_name, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """
    Delete a container device

    name :
        Name of the container

    device_name :
        The device name to delete

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    return _delete_property_dict_item(container, "devices", device_name)


def container_file_put(
    name,
    src,
    dst,
    recursive=False,
    overwrite=False,
    mode=None,
    uid=None,
    gid=None,
    saltenv="base",
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
):
    """
    Put a file into a container

    name :
        Name of the container

    src :
        The source file or directory

    dst :
        The destination file or directory

    recursive :
        Decent into src directory

    overwrite :
        Replace destination if it exists

    mode :
        Set file mode to octal number

    uid :
        Set file uid (owner)

    gid :
        Set file gid (group)

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.container_file_put <container name> /var/tmp/foo /var/tmp/

    """
    # Possibilities:
    #  (src, dst, dir, dir1, and dir2 are directories)
    #  cp /src/file1 /dst/file1
    #  cp /src/file1 /dst/file2
    #  cp /src/file1 /dst
    #  cp /src/file1 /dst/
    #  cp -r /src/dir /dst/
    #  cp -r /src/dir/ /dst/
    #  cp -r /src/dir1 /dst/dir2 (which is not /src/dir1 /dst/dir2/)
    #  cp -r /src/dir1 /dst/dir2/

    # Fix mode. Salt commandline doesn't use octals, so 0600 will be
    # the decimal integer 600 (and not the octal 0600). So, it it's
    # and integer, handle it as if it where a octal representation.
    mode = str(mode)
    if not mode.startswith("0"):
        mode = f"0{mode}"

    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    src = os.path.expanduser(src)

    if not os.path.isabs(src):
        if src.find("://") >= 0:
            cached_file = __salt__["cp.cache_file"](src, saltenv=saltenv)
            if not cached_file:
                raise SaltInvocationError(f"File '{src}' not found")
            if not os.path.isabs(cached_file):
                raise SaltInvocationError("File path must be absolute.")
            src = cached_file

    # Make sure that src doesn't end with '/', unless it's '/'
    src = src.rstrip(os.path.sep)
    if not src:
        src = os.path.sep

    if not os.path.exists(src):
        raise CommandExecutionError(f"No such file or directory '{src}'")

    if os.path.isdir(src) and not recursive:
        raise SaltInvocationError(
            "Cannot copy overwriting a directory without recursive flag set to true!"
        )

    try:
        dst_is_directory = False
        container.files.get(os.path.join(dst, "."))
    except pylxd.exceptions.NotFound:
        pass
    except pylxd.exceptions.LXDAPIException as why:
        if str(why).find("Is a directory") >= 0:
            dst_is_directory = True

    if os.path.isfile(src):
        # Source is a file
        if dst_is_directory:
            dst = os.path.join(dst, os.path.basename(src))
            if not overwrite:
                found = True
                try:
                    container.files.get(os.path.join(dst))
                except pylxd.exceptions.NotFound:
                    found = False
                except pylxd.exceptions.LXDAPIException as why:
                    if str(why).find("not found") >= 0:
                        # Old version of pylxd
                        found = False
                    else:
                        raise
                if found:
                    raise SaltInvocationError(
                        "Destination exists and overwrite is false"
                    )
        if mode is not None or uid is not None or gid is not None:
            # Need to get file stats
            stat = os.stat(src)
            if mode is None:
                mode = oct(stat.st_mode)
            if uid is None:
                uid = stat.st_uid
            if gid is None:
                gid = stat.st_gid

        with salt.utils.files.fopen(src, "rb") as src_fp:
            container.files.put(dst, src_fp.read(), mode=mode, uid=uid, gid=gid)
        return True
    elif not os.path.isdir(src):
        raise SaltInvocationError("Source is neither file nor directory")

    # Source is a directory
    # idx for dstdir = dst + src[idx:]
    if dst.endswith(os.sep):
        idx = len(os.path.dirname(src))
    elif dst_is_directory:
        idx = len(src)
    else:
        # Destination is not a directory and doesn't end with '/'
        # Check that the parent directory of dst exists
        # and is a directory
        try:
            container.files.get(os.path.join(os.path.dirname(dst), "."))
        except pylxd.exceptions.NotFound:
            pass
        except pylxd.exceptions.LXDAPIException as why:
            if str(why).find("Is a directory") >= 0:
                dst_is_directory = True
                # destination is non-existent
                # cp -r /src/dir1 /scr/dir1
                # cp -r /src/dir1 /scr/dir2
                idx = len(src)
                overwrite = True

    # Copy src directory recursive
    if not overwrite:
        raise SaltInvocationError("Destination exists and overwrite is false")

    # Collect all directories first, to create them in one call
    # (for performance reasons)
    dstdirs = []
    for path, _, files in os.walk(src):
        dstdir = os.path.join(dst, path[idx:].lstrip(os.path.sep))
        dstdirs.append(dstdir)
    container.execute(["mkdir", "-p"] + dstdirs)

    set_mode = mode
    set_uid = uid
    set_gid = gid
    # Now transfer the files
    for path, _, files in os.walk(src):
        dstdir = os.path.join(dst, path[idx:].lstrip(os.path.sep))
        for name in files:
            src_name = os.path.join(path, name)
            dst_name = os.path.join(dstdir, name)

            if mode is not None or uid is not None or gid is not None:
                # Need to get file stats
                stat = os.stat(src_name)
                if mode is None:
                    set_mode = oct(stat.st_mode)
                if uid is None:
                    set_uid = stat.st_uid
                if gid is None:
                    set_gid = stat.st_gid

            with salt.utils.files.fopen(src_name, "rb") as src_fp:
                container.files.put(
                    dst_name, src_fp.read(), mode=set_mode, uid=set_uid, gid=set_gid
                )

    return True


def container_file_get(
    name,
    src,
    dst,
    overwrite=False,
    mode=None,
    uid=None,
    gid=None,
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
):
    """
    Get a file from a container

    name :
        Name of the container

    src :
        The source file or directory

    dst :
        The destination file or directory

    mode :
        Set file mode to octal number

    uid :
        Set file uid (owner)

    gid :
        Set file gid (group)

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    """
    # Fix mode. Salt commandline doesn't use octals, so 0600 will be
    # the decimal integer 600 (and not the octal 0600). So, it it's
    # and integer, handle it as if it where a octal representation.

    # Do only if mode is not None, otherwise we get 0None
    if mode is not None:
        mode = str(mode)
        if not mode.startswith("0"):
            mode = f"0{mode}"

    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    dst = os.path.expanduser(dst)
    if not os.path.isabs(dst):
        raise SaltInvocationError("File path must be absolute.")

    if os.path.isdir(dst):
        dst = os.path.join(dst, os.path.basename(src))
    elif not os.path.isdir(os.path.dirname(dst)):
        raise SaltInvocationError("Parent directory for destination doesn't exist.")

    if os.path.exists(dst):
        if not overwrite:
            raise SaltInvocationError("Destination exists and overwrite is false.")
        if not os.path.isfile(dst):
            raise SaltInvocationError("Destination exists but is not a file.")
    else:
        dst_path = os.path.dirname(dst)
        if not os.path.isdir(dst_path):
            raise CommandExecutionError(f"No such file or directory '{dst_path}'")
        # Seems to be duplicate of line 1794, produces /path/file_name/file_name
        # dst = os.path.join(dst, os.path.basename(src))

    with salt.utils.files.fopen(dst, "wb") as df:
        df.write(container.files.get(src))

    if mode:
        os.chmod(dst, mode)
    if uid or uid == "0":
        uid = int(uid)
    else:
        uid = -1
    if gid or gid == "0":
        gid = int(gid)
    else:
        gid = -1
    if uid != -1 or gid != -1:
        os.chown(dst, uid, gid)
    return True


def container_execute(
    name, cmd, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """
    Execute a command list on a container.

    name :
        Name of the container

    cmd :
        Command to be executed (as a list)

        Example :
            '["ls", "-l"]'

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.container_execute <container name> '["ls", "-l"]'

    """
    container = container_get(name, remote_addr, cert, key, verify_cert, _raw=True)
    try:
        result = container.execute(cmd)
        saltresult = {}
        if not hasattr(result, "exit_code"):
            saltresult = dict(
                exit_code=0,
                stdout=result[0],
                stderr=result[1],
            )
        else:
            saltresult = dict(
                exit_code=result.exit_code,
                stdout=result.stdout,
                stderr=result.stderr,
            )
    except pylxd.exceptions.NotFound as e:
        # TODO: Using exit_code 0 here is not always right,
        # in the most cases the command worked ok though.
        # See: https://github.com/lxc/pylxd/issues/280
        saltresult = dict(exit_code=0, stdout="", stderr=str(e))

    if int(saltresult["exit_code"]) > 0:
        saltresult["result"] = False
    else:
        saltresult["result"] = True

    return saltresult


####################
# Profile Management
####################
def profile_list(
    list_names=False, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """Lists all profiles from the LXD.

    list_names :

        Return a list of names instead of full blown dicts.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.profile_list true --out=json
        salt '*' lxd.profile_list --out=json
    """

    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    profiles = client.profiles.all()
    if list_names:
        return [p.name for p in profiles]

    return map(_pylxd_model_to_dict, profiles)


def profile_create(
    name,
    config=None,
    devices=None,
    description=None,
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
):
    """Creates a profile.

    name :
        The name of the profile to get.

    config :
        A config dict or None (None = unset).

        Can also be a list:
            [{'key': 'boot.autostart', 'value': 1},
             {'key': 'security.privileged', 'value': '1'}]

    devices :
        A device dict or None (None = unset).

    description :
        A description string or None (None = unset).

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.profile_create autostart config="{boot.autostart: 1, boot.autostart.delay: 2, boot.autostart.priority: 1}"
        salt '*' lxd.profile_create shared_mounts devices="{shared_mount: {type: 'disk', source: '/home/shared', path: '/home/shared'}}"

    See the `lxd-docs`_ for the details about the config and devices dicts.

    .. _lxd-docs: https://github.com/lxc/lxd/blob/master/doc/rest-api.md#post-10
    """
    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    config, devices = normalize_input_values(config, devices)

    try:
        profile = client.profiles.create(name, config, devices)
    except pylxd.exceptions.LXDAPIException as e:
        raise CommandExecutionError(str(e))

    if description is not None:
        profile.description = description
        pylxd_save_object(profile)

    return _pylxd_model_to_dict(profile)


def profile_get(
    name, remote_addr=None, cert=None, key=None, verify_cert=True, _raw=False
):
    """Gets a profile from the LXD

    name :
        The name of the profile to get.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    _raw :
        Return the pylxd object, this is internal and by states in use.

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.profile_get autostart
    """
    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    profile = None
    try:
        profile = client.profiles.get(name)
    except pylxd.exceptions.LXDAPIException:
        raise SaltInvocationError(f"Profile '{name}' not found")

    if _raw:
        return profile

    return _pylxd_model_to_dict(profile)


def profile_delete(name, remote_addr=None, cert=None, key=None, verify_cert=True):
    """Deletes a profile.

    name :
        The name of the profile to delete.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.profile_delete shared_mounts
    """
    profile = profile_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    profile.delete()
    return True


def profile_config_get(
    name, config_key, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """Get a profile config item.

    name :
        The name of the profile to get the config item from.

    config_key :
        The key for the item to retrieve.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.profile_config_get autostart boot.autostart
    """
    profile = profile_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    return _get_property_dict_item(profile, "config", config_key)


def profile_config_set(
    name,
    config_key,
    config_value,
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
):
    """Set a profile config item.

    name :
        The name of the profile to set the config item to.

    config_key :
        The items key.

    config_value :
        Its items value.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.profile_config_set autostart boot.autostart 0
    """
    profile = profile_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    return _set_property_dict_item(profile, "config", config_key, config_value)


def profile_config_delete(
    name, config_key, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """Delete a profile config item.

    name :
        The name of the profile to delete the config item.

    config_key :
        The config key for the value to retrieve.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.profile_config_delete autostart boot.autostart.delay
    """
    profile = profile_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    return _delete_property_dict_item(profile, "config", config_key)


def profile_device_get(
    name, device_name, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """Get a profile device.

    name :
        The name of the profile to get the device from.

    device_name :
        The name of the device to retrieve.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.profile_device_get default eth0
    """
    profile = profile_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    return _get_property_dict_item(profile, "devices", device_name)


def profile_device_set(
    name,
    device_name,
    device_type="disk",
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
    **kwargs,
):
    """Set a profile device.

    name :
        The name of the profile to set the device to.

    device_name :
        The name of the device to set.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.profile_device_set autostart eth1 nic nictype=bridged parent=lxdbr0
    """
    profile = profile_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    kwargs["type"] = device_type

    for k, v in kwargs.items():
        kwargs[k] = str(v)

    return _set_property_dict_item(profile, "devices", device_name, kwargs)


def profile_device_delete(
    name, device_name, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """Delete a profile device.

    name :
        The name of the profile to delete the device.

    device_name :
        The name of the device to delete.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Example:

    .. code-block:: bash

        salt '*' lxd.profile_device_delete autostart eth1

    """
    profile = profile_get(name, remote_addr, cert, key, verify_cert, _raw=True)

    return _delete_property_dict_item(profile, "devices", device_name)


##################
# Image Management
##################
def image_list(
    list_aliases=False, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """Lists all images from the LXD.

    list_aliases :

        Return a dict with the fingerprint as key and
        a list of aliases as value instead.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.image_list true --out=json
        salt '*' lxd.image_list --out=json
    """
    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    images = client.images.all()
    if list_aliases:
        return {i.fingerprint: [a["name"] for a in i.aliases] for i in images}

    return map(_pylxd_model_to_dict, images)


def image_get(
    fingerprint, remote_addr=None, cert=None, key=None, verify_cert=True, _raw=False
):
    """Get an image by its fingerprint

    fingerprint :
        The fingerprint of the image to retrieve

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    _raw : False
        Return the raw pylxd object or a dict of it?

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.image_get <fingerprint>
    """
    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    image = None
    try:
        image = client.images.get(fingerprint)
    except pylxd.exceptions.LXDAPIException:
        raise SaltInvocationError(f"Image with fingerprint '{fingerprint}' not found")

    if _raw:
        return image

    return _pylxd_model_to_dict(image)


def image_get_by_alias(
    alias, remote_addr=None, cert=None, key=None, verify_cert=True, _raw=False
):
    """Get an image by an alias

    alias :
        The alias of the image to retrieve

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    _raw : False
        Return the raw pylxd object or a dict of it?

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.image_get_by_alias xenial/amd64
    """
    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    image = None
    try:
        image = client.images.get_by_alias(alias)
    except pylxd.exceptions.LXDAPIException:
        raise SaltInvocationError(f"Image with alias '{alias}' not found")

    if _raw:
        return image

    return _pylxd_model_to_dict(image)


def image_delete(image, remote_addr=None, cert=None, key=None, verify_cert=True):
    """Delete an image by an alias or fingerprint

    name :
        The alias or fingerprint of the image to delete,
        can be a obj for the states.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.image_delete xenial/amd64
    """

    image = _verify_image(image, remote_addr, cert, key, verify_cert)

    image.delete()
    return True


def image_from_simplestreams(
    server,
    alias,
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
    aliases=None,
    public=False,
    auto_update=False,
    _raw=False,
):
    """Create an image from simplestreams

    server :
        Simplestreams server URI

    alias :
        The alias of the image to retrieve

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    aliases : []
        List of aliases to append to the copied image

    public : False
        Make this image public available

    auto_update : False
        Should LXD auto update that image?

    _raw : False
        Return the raw pylxd object or a dict of the image?

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.image_from_simplestreams "https://cloud-images.ubuntu.com/releases" "trusty/amd64" aliases='["t", "trusty/amd64"]' auto_update=True
    """
    if aliases is None:
        aliases = []

    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    try:
        image = client.images.create_from_simplestreams(
            server, alias, public=public, auto_update=auto_update
        )
    except pylxd.exceptions.LXDAPIException as e:
        raise CommandExecutionError(str(e))

    # Aliases support
    for alias in aliases:
        image_alias_add(image, alias)

    if _raw:
        return image

    return _pylxd_model_to_dict(image)


def image_from_url(
    url,
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
    aliases=None,
    public=False,
    auto_update=False,
    _raw=False,
):
    """Create an image from an url

    url :
        The URL from where to download the image

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    aliases : []
        List of aliases to append to the copied image

    public : False
        Make this image public available

    auto_update : False
        Should LXD auto update that image?

    _raw : False
        Return the raw pylxd object or a dict of the image?

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.image_from_url https://dl.stgraber.org/lxd aliases='["busybox-amd64"]'
    """
    if aliases is None:
        aliases = []

    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    try:
        image = client.images.create_from_url(
            url, public=public, auto_update=auto_update
        )
    except pylxd.exceptions.LXDAPIException as e:
        raise CommandExecutionError(str(e))

    # Aliases support
    for alias in aliases:
        image_alias_add(image, alias)

    if _raw:
        return image

    return _pylxd_model_to_dict(image)


def image_from_file(
    filename,
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
    aliases=None,
    public=False,
    saltenv="base",
    _raw=False,
):
    """Create an image from a file

    filename :
        The filename of the rootfs

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    aliases : []
        List of aliases to append to the copied image

    public : False
        Make this image public available

    saltenv : base
        The saltenv to use for salt:// copies

    _raw : False
        Return the raw pylxd object or a dict of the image?

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.image_from_file salt://lxd/files/busybox.tar.xz aliases=["busybox-amd64"]
    """
    if aliases is None:
        aliases = []

    cached_file = __salt__["cp.cache_file"](filename, saltenv=saltenv)
    data = b""
    with salt.utils.files.fopen(cached_file, "r+b") as fp:
        data = fp.read()

    client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    try:
        image = client.images.create(data, public=public, wait=True)
    except pylxd.exceptions.LXDAPIException as e:
        raise CommandExecutionError(str(e))

    # Aliases support
    for alias in aliases:
        image_alias_add(image, alias)

    if _raw:
        return image

    return _pylxd_model_to_dict(image)


def image_copy_lxd(
    source,
    src_remote_addr,
    src_cert,
    src_key,
    src_verify_cert,
    remote_addr,
    cert,
    key,
    verify_cert=True,
    aliases=None,
    public=None,
    auto_update=None,
    _raw=False,
):
    """Copy an image from another LXD instance

    source :
        An alias or a fingerprint of the source.

    src_remote_addr :
        An URL to the source remote daemon

        Examples:
            https://mysourceserver.lan:8443

    src_cert :
        PEM Formatted SSL Certificate for the source

        Examples:
            ~/.config/lxc/client.crt

    src_key :
        PEM Formatted SSL Key for the source

        Examples:
            ~/.config/lxc/client.key

    src_verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    remote_addr :
        Address of the destination daemon

        Examples:
            https://mydestserver.lan:8443

    cert :
        PEM Formatted SSL Certificate for the destination

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key for the destination

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    aliases : []
        List of aliases to append to the copied image

    public : None
        Make this image public available, None = copy source

    auto_update : None
        Wherever to auto-update from the original source, None = copy source

    _raw : False
        Return the raw pylxd object or a dict of the destination image?

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.image_copy_lxd xenial/amd64 https://srv01:8443 ~/.config/lxc/client.crt ~/.config/lxc/client.key false https://srv02:8443 ~/.config/lxc/client.crt ~/.config/lxc/client.key false aliases="['xenial/amd64']"
    """
    if aliases is None:
        aliases = []

    log.debug(
        'Trying to copy the image "%s" from "%s" to "%s"',
        source,
        src_remote_addr,
        remote_addr,
    )

    # This will fail with a SaltInvocationError if
    # the image doesn't exists on the source and with a CommandExecutionError
    # on connection problems.
    src_image = None
    try:
        src_image = image_get_by_alias(
            source, src_remote_addr, src_cert, src_key, src_verify_cert, _raw=True
        )
    except SaltInvocationError:
        src_image = image_get(
            source, src_remote_addr, src_cert, src_key, src_verify_cert, _raw=True
        )

    # Will fail with a CommandExecutionError on connection problems.
    dest_client = pylxd_client_get(remote_addr, cert, key, verify_cert)

    dest_image = src_image.copy(
        dest_client, public=public, auto_update=auto_update, wait=True
    )

    # Aliases support
    for alias in aliases:
        image_alias_add(dest_image, alias)

    if _raw:
        return dest_image

    return _pylxd_model_to_dict(dest_image)


def image_alias_add(
    image,
    alias,
    description="",
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
):
    """Create an alias on the given image

    image :
        An image alias, a fingerprint or a image object

    alias :
        The alias to add

    description :
        Description of the alias

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.image_alias_add xenial/amd64 x "Short version of xenial/amd64"
    """
    image = _verify_image(image, remote_addr, cert, key, verify_cert)

    for alias_info in image.aliases:
        if alias_info["name"] == alias:
            return True
    image.add_alias(alias, description)

    return True


def image_alias_delete(
    image, alias, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """Delete an alias (this is currently not restricted to the image)

    image :
        An image alias, a fingerprint or a image object

    alias :
        The alias to delete

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if
        you provide remote_addr and its a TCP Address!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.image_alias_add xenial/amd64 x "Short version of xenial/amd64"
    """
    image = _verify_image(image, remote_addr, cert, key, verify_cert)

    try:
        image.delete_alias(alias)
    except pylxd.exceptions.LXDAPIException:
        return False

    return True


#####################
# Snapshot Management
#####################


def snapshots_all(container, remote_addr=None, cert=None, key=None, verify_cert=True):
    """
    Get all snapshots for a container

    container :
        The name of the container to get.

    remote_addr :
        An URL to a remote server. The 'cert' and 'key' fields must also be
        provided if 'remote_addr' is defined.

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Verify the ssl certificate.  Default: True

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.snapshots_all test-container
    """
    containers = container_get(
        container, remote_addr, cert, key, verify_cert, _raw=True
    )
    if container:
        containers = [containers]
    ret = {}
    for cont in containers:
        ret.update({cont.name: [{"name": c.name} for c in cont.snapshots.all()]})

    return ret


def snapshots_create(
    container, name=None, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """
    Create a snapshot for a container

    container :
        The name of the container to get.

    name :
        The name of the snapshot.

    remote_addr :
        An URL to a remote server. The 'cert' and 'key' fields must also be
        provided if 'remote_addr' is defined.

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Verify the ssl certificate.  Default: True

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.snapshots_create test-container test-snapshot
    """
    cont = container_get(container, remote_addr, cert, key, verify_cert, _raw=True)
    if not name:
        name = datetime.now().strftime("%Y%m%d%H%M%S")

    cont.snapshots.create(name)

    for c in snapshots_all(container).get(container):
        if c.get("name") == name:
            return {"name": name}

    return {"name": False}


def snapshots_delete(
    container, name, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """
    Delete a snapshot for a container

    container :
        The name of the container to get.

    name :
        The name of the snapshot.

    remote_addr :
        An URL to a remote server. The 'cert' and 'key' fields must also be
        provided if 'remote_addr' is defined.

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Verify the ssl certificate.  Default: True

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.snapshots_delete test-container test-snapshot
    """
    cont = container_get(container, remote_addr, cert, key, verify_cert, _raw=True)

    try:
        for s in cont.snapshots.all():
            if s.name == name:
                s.delete()
                return True
    except pylxd.exceptions.LXDAPIException:
        pass

    return False


def snapshots_get(
    container, name, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """
    Get information about snapshot for a container

    container :
        The name of the container to get.

    name :
        The name of the snapshot.

    remote_addr :
        An URL to a remote server. The 'cert' and 'key' fields must also be
        provided if 'remote_addr' is defined.

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Certificate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Verify the ssl certificate.  Default: True

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.snapshots_get test-container test-snapshot
    """
    container = container_get(container, remote_addr, cert, key, verify_cert, _raw=True)
    return container.snapshots.get(name)


################
# Helper Methods
################


def normalize_input_values(config, devices):
    """
    normalize config input so returns can be put into mongodb, which doesn't like `.`

    This is not meant to be used on the commandline.

    CLI Examples:

    .. code-block:: bash

        salt '*' lxd.normalize_input_values config={} devices={}
    """
    if isinstance(config, list):
        if config and "key" in config[0] and "value" in config[0]:
            config = {d["key"]: d["value"] for d in config}
        else:
            config = {}

    if isinstance(config, str):
        raise SaltInvocationError("config can't be a string, validate your YAML input.")

    if isinstance(devices, str):
        raise SaltInvocationError(
            "devices can't be a string, validate your YAML input."
        )

    # Golangs wants strings
    if config is not None:
        for k, v in config.items():
            config[k] = str(v)
    if devices is not None:
        for dn in devices:
            for k, v in devices[dn].items():
                devices[dn][k] = v

    return (
        config,
        devices,
    )


def sync_config_devices(obj, newconfig, newdevices, test=False):
    """Syncs the given config and devices with the object
    (a profile or a container)
    returns a changes dict with all changes made.

    obj :
        The object to sync with / or just test with.

    newconfig:
        The new config to check with the obj.

    newdevices:
        The new devices to check with the obj.

    test:
        Wherever to not change anything and give "Would change" message.
    """
    changes = {}

    #
    # config changes
    #
    if newconfig is None:
        newconfig = {}

    newconfig = dict(
        list(zip(map(str, newconfig.keys()), map(str, newconfig.values())))
    )
    cck = set(newconfig.keys())

    obj.config = dict(
        list(zip(map(str, obj.config.keys()), map(str, obj.config.values())))
    )
    ock = set(obj.config.keys())

    config_changes = {}
    # Removed keys
    for k in ock.difference(cck):
        # Ignore LXD internals.
        if k.startswith("volatile.") or k.startswith("image."):
            continue

        if not test:
            config_changes[k] = 'Removed config key "{}", its value was "{}"'.format(
                k, obj.config[k]
            )
            del obj.config[k]
        else:
            config_changes[k] = 'Would remove config key "{} with value "{}"'.format(
                k, obj.config[k]
            )

    # same keys
    for k in cck.intersection(ock):
        # Ignore LXD internals.
        if k.startswith("volatile.") or k.startswith("image."):
            continue

        if newconfig[k] != obj.config[k]:
            if not test:
                config_changes[k] = (
                    'Changed config key "{}" to "{}", its value was "{}"'.format(
                        k, newconfig[k], obj.config[k]
                    )
                )
                obj.config[k] = newconfig[k]
            else:
                config_changes[k] = (
                    'Would change config key "{}" to "{}", its current value is "{}"'.format(
                        k, newconfig[k], obj.config[k]
                    )
                )

    # New keys
    for k in cck.difference(ock):
        # Ignore LXD internals.
        if k.startswith("volatile.") or k.startswith("image."):
            continue

        if not test:
            config_changes[k] = f'Added config key "{k}" = "{newconfig[k]}"'
            obj.config[k] = newconfig[k]
        else:
            config_changes[k] = 'Would add config key "{}" = "{}"'.format(
                k, newconfig[k]
            )

    if config_changes:
        changes["config"] = config_changes

    #
    # devices changes
    #
    if newdevices is None:
        newdevices = {}

    dk = set(obj.devices.keys())
    ndk = set(newdevices.keys())

    devices_changes = {}
    # Removed devices
    for k in dk.difference(ndk):
        # Ignore LXD internals.
        if k == "root":
            continue

        if not test:
            devices_changes[k] = f'Removed device "{k}"'
            del obj.devices[k]
        else:
            devices_changes[k] = f'Would remove device "{k}"'

    # Changed devices
    for k, v in obj.devices.items():
        # Ignore LXD internals also for new devices.
        if k == "root":
            continue

        if k not in newdevices:
            # In test mode we don't delete devices above.
            continue

        if newdevices[k] != v:
            if not test:
                devices_changes[k] = f'Changed device "{k}"'
                obj.devices[k] = newdevices[k]
            else:
                devices_changes[k] = f'Would change device "{k}"'

    # New devices
    for k in ndk.difference(dk):
        # Ignore LXD internals.
        if k == "root":
            continue

        if not test:
            devices_changes[k] = f'Added device "{k}"'
            obj.devices[k] = newdevices[k]
        else:
            devices_changes[k] = f'Would add device "{k}"'

    if devices_changes:
        changes["devices"] = devices_changes

    return changes


def _set_property_dict_item(obj, prop, key, value):
    """Sets the dict item key of the attr from obj.

    Basicaly it does getattr(obj, prop)[key] = value.


    For the disk device we added some checks to make
    device changes on the CLI saver.
    """
    attr = getattr(obj, prop)
    if prop == "devices":
        device_type = value["type"]

        if device_type == "disk":

            if "path" not in value:
                raise SaltInvocationError("path must be given as parameter")

            if value["path"] != "/" and "source" not in value:
                raise SaltInvocationError("source must be given as parameter")

        for k in value.keys():
            if k.startswith("__"):
                del value[k]

        attr[key] = value

    else:  # config
        attr[key] = str(value)

    pylxd_save_object(obj)

    return _pylxd_model_to_dict(obj)


def _get_property_dict_item(obj, prop, key):
    attr = getattr(obj, prop)
    if key not in attr:
        raise SaltInvocationError(f"'{key}' doesn't exists")

    return attr[key]


def _delete_property_dict_item(obj, prop, key):
    attr = getattr(obj, prop)
    if key not in attr:
        raise SaltInvocationError(f"'{key}' doesn't exists")

    del attr[key]
    pylxd_save_object(obj)

    return True


def _verify_image(image, remote_addr=None, cert=None, key=None, verify_cert=True):
    # Get image by alias/fingerprint or check for fingerprint attribute
    if isinstance(image, str):
        name = image

        # This will fail with a SaltInvocationError if
        # the image doesn't exists on the source and with a
        # CommandExecutionError on connection problems.
        image = None
        try:
            image = image_get_by_alias(
                name, remote_addr, cert, key, verify_cert, _raw=True
            )
        except SaltInvocationError:
            image = image_get(name, remote_addr, cert, key, verify_cert, _raw=True)
    elif not hasattr(image, "fingerprint"):
        raise SaltInvocationError(f"Invalid image '{image}'")
    return image


def _pylxd_model_to_dict(obj):
    """Translates a plyxd model object to a dict"""
    marshalled = {}
    for key in obj.__attributes__.keys():
        if hasattr(obj, key):
            marshalled[key] = getattr(obj, key)
    return marshalled


#
# Monkey patching for missing functionality in pylxd
#

if HAS_PYLXD:
    import pylxd.exceptions  # NOQA

    if not hasattr(pylxd.exceptions, "NotFound"):
        # Old version of pylxd

        class NotFound(pylxd.exceptions.LXDAPIException):
            """An exception raised when an object is not found."""

        pylxd.exceptions.NotFound = NotFound

    try:
        from pylxd.container import Container
    except ImportError:
        from pylxd.models.container import Container

    def put(self, filepath, data, mode=None, uid=None, gid=None):
        headers = {}
        if mode is not None:
            if isinstance(mode, int):
                mode = oct(mode)
            elif not mode.startswith("0"):
                mode = f"0{mode}"
            headers["X-LXD-mode"] = mode
        if uid is not None:
            headers["X-LXD-uid"] = str(uid)
        if gid is not None:
            headers["X-LXD-gid"] = str(gid)
        response = self._client.api.containers[self._container.name].files.post(
            params={"path": filepath}, data=data, headers=headers
        )
        return response.status_code == 200

    Container.FilesManager.put = put



import xml.etree.ElementTree as ET

import salt.utils.xmlutil as xmlutil


def append_to_XMLDesc(mocked, fragment):
    """
    Append an XML fragment at the end of the mocked XMLDesc return_value of mocked.
    """
    xml_doc = ET.fromstring(mocked.XMLDesc())
    xml_fragment = ET.fromstring(fragment)
    xml_doc.append(xml_fragment)
    mocked.XMLDesc.return_value = ET.tostring(xml_doc)


def assert_xml_equals(actual, expected):
    """
    Assert that two ElementTree nodes are equal
    """
    assert xmlutil.to_dict(xmlutil.strip_spaces(actual), True) == xmlutil.to_dict(
        xmlutil.strip_spaces(expected), True
    )


def strip_xml(xml_str):
    """
    Remove all spaces and formatting from an XML string
    """
    return ET.tostring(xmlutil.strip_spaces(ET.fromstring(xml_str)))


def assert_called(mock, condition):
    """
    Assert that the mock has been called if not in test mode, and vice-versa.
    I know it's a simple XOR, but makes the tests easier to read
    """
    assert not condition and not mock.called or condition and mock.called


def assert_equal_unit(actual, expected, unit="KiB"):
    """
    Assert that two ElementTree nodes have the same value and unit
    """
    assert actual.get("unit") == unit
    assert actual.text == str(expected)



"""
Helpful generators and other tools
"""

import fnmatch
import re

import salt.utils.files


def split(orig, sep=None):
    """
    Generator function for iterating through large strings, particularly useful
    as a replacement for str.splitlines().

    See http://stackoverflow.com/a/3865367
    """
    exp = re.compile(r"\s+" if sep is None else re.escape(sep))
    pos = 0
    length = len(orig)
    while True:
        match = exp.search(orig, pos)
        if not match:
            if pos < length or sep is not None:
                val = orig[pos:]
                if val:
                    # Only yield a value if the slice was not an empty string,
                    # because if it is then we've reached the end. This keeps
                    # us from yielding an extra blank value at the end.
                    yield val
            break
        if pos < match.start() or sep is not None:
            yield orig[pos : match.start()]
        pos = match.end()


def read_file(fh_, chunk_size=1048576):
    """
    Generator that reads chunk_size bytes at a time from a file/filehandle and
    yields it.
    """
    try:
        if chunk_size != int(chunk_size):
            raise ValueError
    except ValueError:
        raise ValueError("chunk_size must be an integer")
    try:
        while True:
            try:
                chunk = fh_.read(chunk_size)
            except AttributeError:
                # Open the file and re-attempt the read
                fh_ = salt.utils.files.fopen(fh_, "rb")  # pylint: disable=W8470
                chunk = fh_.read(chunk_size)
            if not chunk:
                break
            yield chunk
    finally:
        try:
            fh_.close()
        except AttributeError:
            pass


def fnmatch_multiple(candidates, pattern):
    """
    Convenience function which runs fnmatch.fnmatch() on each element of passed
    iterable. The first matching candidate is returned, or None if there is no
    matching candidate.
    """
    # Make sure that candidates is iterable to avoid a TypeError when we try to
    # iterate over its items.
    try:
        candidates_iter = iter(candidates)
    except TypeError:
        return None

    for candidate in candidates_iter:
        try:
            if fnmatch.fnmatch(candidate, pattern):
                return candidate
        except TypeError:
            pass
    return None



import pytest

import salt.modules.nexus as nexus
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {nexus: {}}


def test_artifact_get_metadata():
    with patch(
        "salt.modules.nexus._get_artifact_metadata_xml",
        MagicMock(
            return_value="""<?xml version="1.0" encoding="UTF-8"?>
<metadata>
  <groupId>com.company.sampleapp.web-module</groupId>
  <artifactId>web</artifactId>
  <versioning>
    <release>0.1.0</release>
    <versions>
      <version>0.0.1</version>
      <version>0.0.2</version>
      <version>0.0.3</version>
      <version>0.1.0</version>
    </versions>
    <lastUpdated>20171010143552</lastUpdated>
  </versioning>
</metadata>"""
        ),
    ):
        metadata = nexus._get_artifact_metadata(
            nexus_url="http://nexus.example.com/repository",
            repository="libs-releases",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            headers={},
        )
        assert metadata["latest_version"] == "0.1.0"


def test_snapshot_version_get_metadata():
    with patch(
        "salt.modules.nexus._get_snapshot_version_metadata_xml",
        MagicMock(
            return_value="""<?xml version="1.0" encoding="UTF-8"?>
<metadata modelVersion="1.1.0">
  <groupId>com.company.sampleapp.web-module</groupId>
  <artifactId>web</artifactId>
  <version>0.0.2-SNAPSHOT</version>
  <versioning>
    <snapshot>
      <timestamp>20170920.212353</timestamp>
      <buildNumber>3</buildNumber>
    </snapshot>
    <lastUpdated>20171112171500</lastUpdated>
    <snapshotVersions>
      <snapshotVersion>
        <classifier>sans-externalized</classifier>
        <extension>jar</extension>
        <value>0.0.2-20170920.212353-3</value>
        <updated>20170920212353</updated>
      </snapshotVersion>
      <snapshotVersion>
        <classifier>dist</classifier>
        <extension>zip</extension>
        <value>0.0.2-20170920.212353-3</value>
        <updated>20170920212353</updated>
      </snapshotVersion>
    </snapshotVersions>
  </versioning>
</metadata>"""
        ),
    ):
        metadata = nexus._get_snapshot_version_metadata(
            nexus_url="http://nexus.example.com/repository",
            repository="libs-releases",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            version="0.0.2-SNAPSHOT",
            headers={},
        )
        assert metadata["snapshot_versions"]["dist"] == "0.0.2-20170920.212353-3"


def test_artifact_metadata_url():
    metadata_url = nexus._get_artifact_metadata_url(
        nexus_url="http://nexus.example.com/repository",
        repository="libs-releases",
        group_id="com.company.sampleapp.web-module",
        artifact_id="web",
    )

    assert (
        metadata_url
        == "http://nexus.example.com/repository/libs-releases/com/company/sampleapp/web-module/web/maven-metadata.xml"
    )


def test_snapshot_version_metadata_url():
    metadata_url = nexus._get_snapshot_version_metadata_url(
        nexus_url="http://nexus.example.com/repository",
        repository="libs-snapshots",
        group_id="com.company.sampleapp.web-module",
        artifact_id="web",
        version="0.0.2-SNAPSHOT",
    )

    assert (
        metadata_url
        == "http://nexus.example.com/repository/libs-snapshots/com/company/sampleapp/web-module/web/0.0.2-SNAPSHOT/maven-metadata.xml"
    )


def test_construct_url_for_released_version():
    artifact_url, file_name = nexus._get_release_url(
        repository="libs-releases",
        group_id="com.company.sampleapp.web-module",
        artifact_id="web",
        packaging="zip",
        nexus_url="http://nexus.example.com/repository",
        version="0.1.0",
    )

    assert (
        artifact_url
        == "http://nexus.example.com/repository/libs-releases/com/company/sampleapp/web-module/web/0.1.0/web-0.1.0.zip"
    )
    assert file_name == "web-0.1.0.zip"


def test_construct_url_for_snapshot_version():
    with patch(
        "salt.modules.nexus._get_snapshot_version_metadata",
        MagicMock(
            return_value={"snapshot_versions": {"zip": "0.0.2-20170920.212353-3"}}
        ),
    ):

        artifact_url, file_name = nexus._get_snapshot_url(
            nexus_url="http://nexus.example.com/repository",
            repository="libs-snapshots",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            version="0.2.0-SNAPSHOT",
            packaging="zip",
            headers={},
        )

        assert (
            artifact_url
            == "http://nexus.example.com/repository/libs-snapshots/com/company/sampleapp/web-module/web/0.2.0-SNAPSHOT/web-0.0.2-20170920.212353-3.zip"
        )
        assert file_name == "web-0.0.2-20170920.212353-3.zip"



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.memcached as memcached
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class MemcachedTestCase(TestCase):
    """
    Test cases for salt.modules.memcached
    """

    # 'status' function tests: 2

    def test_status(self):
        """
        Test if it gets memcached status
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertDictEqual(memcached.status(), {"127.0.0.1:11211 (1)": {}})

    def test_status_false(self):
        """
        Test if it gets memcached status
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return []

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertFalse(memcached.status())

    # 'get' function tests: 1

    def test_get(self):
        """
        Test if it retrieve value for a key
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

            @staticmethod
            def get(key):
                """
                Mock of get method
                """
                return key

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertEqual(memcached.get("salt"), "salt")

    # 'set_' function tests: 1

    def test_set(self):
        """
        Test if it set a key on the memcached server
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            def __init__(self):
                self.key = None
                self.value = None
                self.time = None
                self.min_compress_len = None

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

            def set(self, key, value, time, min_compress_len):
                """
                Mock of set method
                """
                self.key = key
                self.value = value
                self.time = time
                self.min_compress_len = min_compress_len
                return True

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertTrue(memcached.set_("salt", "1111"))

            self.assertRaises(
                SaltInvocationError, memcached.set_, "salt", "1111", time="0.1"
            )

            self.assertRaises(
                SaltInvocationError,
                memcached.set_,
                "salt",
                "1111",
                min_compress_len="0.1",
            )

    # 'delete' function tests: 1

    def test_delete(self):
        """
        Test if it delete a key from memcache server
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            def __init__(self):
                self.key = None
                self.time = None

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

            def delete(self, key, time):
                """
                Mock of delete method
                """
                self.key = key
                self.time = time
                return True

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertTrue(memcached.delete("salt"))

            self.assertRaises(
                SaltInvocationError, memcached.delete, "salt", "1111", time="0.1"
            )

    # 'add' function tests: 1

    def test_add(self):
        """
        Test if it add a key from memcache server
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            def __init__(self):
                self.key = None
                self.value = None
                self.time = None
                self.min_compress_len = None

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

            def add(self, key, value, time, min_compress_len):
                """
                Mock of add method
                """
                self.key = key
                self.value = value
                self.time = time
                self.min_compress_len = min_compress_len
                return True

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertTrue(memcached.add("salt", "1111"))

            self.assertRaises(
                SaltInvocationError, memcached.add, "salt", "1111", time="0.1"
            )

            self.assertRaises(
                SaltInvocationError,
                memcached.add,
                "salt",
                "1111",
                min_compress_len="0.1",
            )

    # 'replace' function tests: 1

    def test_replace(self):
        """
        Test if it replace a key from memcache server
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            def __init__(self):
                self.key = None
                self.value = None
                self.time = None
                self.min_compress_len = None

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

            def replace(self, key, value, time, min_compress_len):
                """
                Mock of replace method
                """
                self.key = key
                self.value = value
                self.time = time
                self.min_compress_len = min_compress_len
                return True

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertTrue(memcached.replace("salt", "1111"))

            self.assertRaises(
                SaltInvocationError, memcached.replace, "salt", "1111", time="0.1"
            )

            self.assertRaises(
                SaltInvocationError,
                memcached.replace,
                "salt",
                "1111",
                min_compress_len="0.1",
            )

    # 'increment' function tests: 3

    def test_increment(self):
        """
        Test if it increment the value of a key
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            def __init__(self):
                self.key = None

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

            def get(self, key):
                """
                Mock of get method
                """
                self.key = key
                return 1

            def incr(self, key, delta):
                """
                Mock of incr method
                """
                self.key = key
                if not isinstance(delta, int):
                    raise SaltInvocationError("Delta value must be an integer")
                return key

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertEqual(memcached.increment("salt"), "salt")

            self.assertRaises(
                SaltInvocationError, memcached.increment, "salt", delta="sa"
            )

    def test_increment_exist(self):
        """
        Test if it increment the value of a key
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            def __init__(self):
                self.key = None

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

            def get(self, key):
                """
                Mock of get method
                """
                self.key = key
                return key

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertRaises(CommandExecutionError, memcached.increment, "salt")

    def test_increment_none(self):
        """
        Test if it increment the value of a key
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            def __init__(self):
                self.key = None

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

            def get(self, key):
                """
                Mock of get method
                """
                self.key = key
                return None

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertRaises(CommandExecutionError, memcached.increment, "salt")

    # 'decrement' function tests: 3

    def test_decrement(self):
        """
        Test if it decrement the value of a key
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            def __init__(self):
                self.key = None

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

            def get(self, key):
                """
                Mock of get method
                """
                self.key = key
                return 1

            def decr(self, key, delta):
                """
                Mock of decr method
                """
                self.key = key
                if not isinstance(delta, int):
                    raise SaltInvocationError("Delta value must be an integer")
                return key

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertEqual(memcached.decrement("salt"), "salt")

            self.assertRaises(
                SaltInvocationError, memcached.decrement, "salt", delta="sa"
            )

    def test_decrement_exist(self):
        """
        Test if it decrement the value of a key
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            def __init__(self):
                self.key = None

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

            def get(self, key):
                """
                Mock of get method
                """
                self.key = key
                return key

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertRaises(CommandExecutionError, memcached.decrement, "salt")

    def test_decrement_none(self):
        """
        Test if it decrement the value of a key
        """

        class MockMemcache:
            """
            Mock of memcache
            """

            def __init__(self):
                self.key = None

            @staticmethod
            def get_stats():
                """
                Mock of stats method
                """
                return [("127.0.0.1:11211 (1)", {})]

            def get(self, key):
                """
                Mock of get method
                """
                self.key = key
                return None

        with patch.object(
            memcached, "_connect", MagicMock(return_value=MockMemcache())
        ):
            self.assertRaises(CommandExecutionError, memcached.decrement, "salt")
